#pragma once

#include <lua.hpp>

namespace Bix
{


	class State {
	public:
		lua_State *_l;
	public:
		State() : _l(nullptr) { _l = luaL_newstate(); }
		State(const State &other) = delete;
		State &operator=(const State &other) = delete;
		State(State &&other) : _l(other._l) { other._l = nullptr; }
		~State() { if (_l == nullptr) return; lua_close(_l); }

		template<typename T> void Push(const T value);
		template<typename T> T Read(int index) const;

		template <typename T, typename... Ts>
		int Push(const T value, const Ts... values) {
			Push<T>(value);
			Push(values...);
			return sizeof...(values) + 1;
		}

		void Push(){}

		

		template <size_t, typename... Ts>
		struct _pop {
			typedef std::tuple<Ts...> type;

			// base case: creates a tuple containing one element
			template <typename T>
			static std::tuple<T> worker(const State &s, const int index) {
				return std::make_tuple(s.Read<T>(index));
			}

			// inductive case
			template <typename T1, typename T2, typename... Rest>
			static std::tuple<T1, T2, Rest...> worker(const State &s,
				const int index) {
				std::tuple<T1> head = std::make_tuple(s.Read<T1>(index));
				return std::tuple_cat(head, worker<T2, Rest...>(s, index + 1));
			}

			static type apply(State &s) {
				auto ret = worker<Ts...>(s, 1);
				lua_pop(s._l, (int)sizeof...(Ts));
				return ret;
			}
		};


		
		// Specialization for singular type
		template <typename T>
		struct _pop<1, T> {
			typedef T type;
			static type apply(State &s)
			{
				// Read the top element (negative indices count from the top)
				type ret = s.Read<T>(-1);

				// Remove it from the stack
				lua_pop(s._l, 1);
				return ret;
			}
		};

		// Specialization for no return
		template <typename... Ts>
		struct _pop<0, Ts...> {
			typedef void type;
			static type apply(State &s) {}
		};


		template <typename... T>
		typename _pop<sizeof...(T), T...>::type Pop() {
			return _pop<sizeof...(T), T...>::apply(*this);
		}



		template <typename... Ret, typename... Args>
		typename _pop<sizeof...(Ret),Ret...>::type Call(const std::string &fun,
			const Args&... args) {
			// Push function to stack
			lua_getglobal(_l, fun.c_str());

			const int num_args = sizeof...(Args);
			const int num_ret = sizeof...(Ret);

			// Push all arguments to the stack using our variadic Push
			Push(args...);

			// Call the function
			lua_call(_l, num_args, num_ret);

			// Return all the results and remove the from the stack
			return Pop<Ret...>();
		}

	};

	template<> void State::Push(const bool value)			{ lua_pushboolean(_l, value); }
	template<> void State::Push(const int value)			{ lua_pushinteger(_l, value); }
	template<> void State::Push(const float value)			{ lua_pushnumber(_l, value); }
	template<> void State::Push(const double value)			{ lua_pushnumber(_l, value); }
	template<> void State::Push(const std::string value)	{ lua_pushstring(_l, value.c_str()); }
	template<> void State::Push(const char *value)			{ lua_pushstring(_l, value); }

	template<> bool			State::Read(int index) const	{ return (lua_toboolean(_l, index)>0) ? true : false; }
	template<> int			State::Read(int index) const	{ return lua_tointeger(_l, index); }
	template<> float		State::Read(int index) const	{ return (float)lua_tonumber(_l, index); }
	template<> double		State::Read(int index) const	{ return lua_tonumber(_l, index); }
	template<> std::string	State::Read(int index) const	{ return std::string(lua_tostring(_l, index)); }
	template<> const char*	State::Read(int index) const	{ return lua_tostring(_l, index); }



}
